import * as ini from "ini";
import * as  fs from "fs";

type IGenericField<T, Co> = Co | {
   type: Co;
   default?: T;
   optional?: boolean;
}

type IValidTypes = string | number | boolean;

type ITestField<T1 extends IValidTypes, T2 extends IValidTypes, R> = T1 extends T2 ? R : never;

export type IField<F extends IValidTypes> =
   ITestField<F, string, IGenericField<string, StringConstructor>> |
   ITestField<F, number, IGenericField<number, NumberConstructor>> |
   ITestField<F, boolean, IGenericField<boolean, BooleanConstructor>>

export type IConfigInterface = {
   [category: string]: {
      [field: string]: IValidTypes
   }
}

export type IConfig<T extends IConfigInterface> = {
   [category in keyof T]: {
      [field in keyof T[category]]: IField<T[category][field]>
   }
}

function parseData(config: IConfig<any>, rawdata: string) {
   const data = rawdata ? ini.parse(rawdata) : {};
   let result: any = {};
   let errors: string[] = [];
   for (let C in config) {
      result[C] = {};
      for (let F in config[C]) {
         let gfieldConfig = config[C][F];
         let fieldConfig: Exclude<typeof gfieldConfig, NumberConstructor | StringConstructor | BooleanConstructor> = ((gfieldConfig === Number || gfieldConfig === String || gfieldConfig === Boolean) ? { type: gfieldConfig } : gfieldConfig) as any;

         const envname = `${C.toUpperCase()}_${F.toUpperCase()}`;
         let value = (data[C] || {})[F] || process.env[envname] || fieldConfig.default;

         if ((value === undefined || value === null) && !fieldConfig.optional) {
            errors.push(`[${C}][${F}] Cannot be undefined!`)
         } else {
            switch (fieldConfig.type) {
               case Number:
                  value = Number(value)
                  break;
               case String:
                  value = String(value);
                  break;
               case Boolean:
                  value = Boolean(value);
                  break;
            }
            result[C][F] = value
         }
      }
   }

   if (errors.length > 0)
      throw new Error("There are errors: \n" + errors.join("\n"));

   return result;
}

function parse<T extends IConfigInterface>(config: IConfig<T>, path?: string) {
   let data;
   if (path)
      data = fs.existsSync(path) ? fs.readFileSync(path).toString("utf8") : undefined;
   return parseData(config, data) as T;
}

async function parseAsync<T extends IConfigInterface>(config: IConfig<T>, path: string) {
   let data;
   if (path && await new Promise(yes => fs.exists(path, yes))) {
      data = await new Promise<Buffer>((yes, no) => fs.readFile(path, (err, data) => err ? no(err) : yes(data))).then(res => res.toString("utf8"));
   }
   return parseData(config, data) as T;
}

export {
   parse,
   parseAsync
}